# Imports
import numpy as np
import warnings

# Edge Dict
from .edges import EdgeDict

def XES(mcp_data):
    """Sum the MCP detector image over all recorded datapoints."""
    try:
        sum = np.sum(mcp_data, axis=1)

    except:
        warnings.warn("Only one data point in MCP file to sum.")
        sum = mcp_data

    return sum

def rXES(mono_energy, mcp_data_norm, xes_incident_start, xes_incident_end):
    """Calculate resonant emission at selected energy (MCP)."""
    idx_start = (np.abs(xes_incident_start - mono_energy)).argmin()
    idx_end = (np.abs(xes_incident_end - mono_energy)).argmin()

    return mcp_data_norm[idx_start:idx_end, :].sum(axis=0)

def XRF(sdd_data):
    """Sum the SDD detector image over all recorded datapoints"""
    return np.sum(sdd_data, axis=0)

def rXRF(mono_energy, sdd_data_norm, xes_incident_start, xes_incident_end):
    """Calculate resonant emission at selected energy (SDD)."""
    idx_start = (np.abs(xes_incident_start - mono_energy)).argmin()
    idx_end = (np.abs(xes_incident_end - mono_energy)).argmin()

    return sdd_data_norm[idx_start:idx_end, :].sum(axis=0)

def ROI(PFY_edge=None, iPFY_edge=None):
    """Set a region of interest (ROI) over which to sum PFY (and iPFY)

        Parameters:
            PFY_edge : specify the absorption egde over which to sum.
                        Options:
                            N, O

            iPFY_edge : iPFY_edge when setting an ROI for iPFY
    """

    if PFY_edge != None:
        SDDLowerBound = EdgeDict[PFY_edge][0]
        SDDUpperBound = EdgeDict[PFY_edge][1]

        return SDDLowerBound, SDDUpperBound

    elif iPFY_edge != None:
        iSDDLowerBound = EdgeDict[iPFY_edge][0]
        iSDDUpperBound = EdgeDict[iPFY_edge][1]

        return iSDDLowerBound, iSDDUpperBound

    else:
        raise Exception("Proper ROI bounds not defined.")

def PFY(sdd_energy, sdd_data, mesh_current, PFY_edge=None, SDDLowerBound=None, SDDUpperBound=None):
    """ Calculate PFY based on a defined ROI.

    Prerequisites:
        Option 1: Define ROI via ROI method
        Option 2: Define ROI explictily by calling
                    SDDLowerBound : lower energy limit for summation
                    SDDUpperBound : upper energy limit for summation

    """
    if PFY_edge != None:
        SDDLowerBound, SDDUpperBound = ROI(PFY_edge)

    elif SDDLowerBound != None and SDDUpperBound != None:
        pass

    else:
        raise Exception(
            "Proper PFY boundaries not defined. Define boundaries or use TFY.")

    idx_low_sdd = (np.abs(SDDLowerBound - sdd_energy)).argmin()
    idx_high_sdd = (np.abs(SDDUpperBound - sdd_energy)).argmin()

    # Transform this to PFY spectrum (rows = incident energy points, columns = detector 'space')
    sdd_detector_sum = sdd_data[:,
                                    idx_low_sdd:idx_high_sdd].sum(axis=1)
    PFY_spec = (sdd_detector_sum/mesh_current)

    return PFY_spec

def iPFY(sdd_energy, sdd_data, mesh_current, iPFY_edge=None, iSDDLowerBound=None, iSDDUpperBound=None):
    """ Calculate iPFY based on a defined ROI.

        Prerequisites:
            Option 1: Define ROI via ROI method (and specify **kwarg iPFY_edge)
            Option 2: Define ROI explictily by calling
                    iSDDLowerBound : lower energy limit for summation
                    iSDDUpperBound : upper energy limit for summation

    """
    if iPFY_edge != None:
        iSDDLowerBound, iSDDUpperBound = ROI(iPFY_edge)

    elif iSDDLowerBound != None and iSDDUpperBound != None:
        pass

    else:
        raise Exception("Proper iPFY boundaries not defined.")

    iidx_low_sdd = (np.abs(iSDDLowerBound - sdd_energy)).argmin()
    iidx_high_sdd = (np.abs(iSDDUpperBound - sdd_energy)).argmin()
    sdd_detector_sum_i = sdd_data[:,
                                        iidx_low_sdd:iidx_high_sdd].sum(axis=1)
    iPFY_spec = 1/(sdd_detector_sum_i/mesh_current)

    return iPFY_spec

def TFY(sdd_data, mesh_current):
    """Calculate TFY by summing over entire SDD image without ROI."""
    # Transform this to TFY spectrum (rows = incident energy points, columns = detector 'space')
    sdd_detector_sum = sdd_data[:, :].sum(axis=1)
    TFY_spec = (sdd_detector_sum/mesh_current)

    return TFY_spec

def specPFY(mcp_energy, mcp_data_norm, mcp_lowE, mcp_highE):
    """Calculate spectrometer PFY based on ROI set."""
    mcp_highE_idx = (np.abs(mcp_highE-mcp_energy)).argmin()
    mcp_lowE_idx = (np.abs(mcp_lowE-mcp_energy)).argmin()

    return mcp_data_norm[:, mcp_lowE_idx:mcp_highE_idx].sum(axis=1)


from .util import doesMatchPattern, check_key_in_dict
from .ReadData import REIXS
from .edges import EdgeDict
from .xeol import *
from .simplemath import apply_offset, take_derivative1d
import warnings
import numpy as np
from .parser import math_stream
from scipy.interpolate import interp1d


def loadSCAscans(basedir, file, x_stream, y_stream, *args, norm=True, is_XAS=False, xoffset=None, xcoffset=None, yoffset=None, ycoffset=None, background=None, deriv=None, energyloss=None, grid_x=[None, None, None]):
    special_streams = ['TEY', 'TFY', 'PFY', 'iPFY', 'XES', 'rXES', 'specPFY',
                       'XRF', 'rXRF', 'XEOL', 'rXEOL', 'POY', 'TOY', 'EY', 'Sample', 'Mesh']  # all special inputs
    XAS_streams = ['TEY', 'TFY', 'PFY', 'iPFY', 'specPFY', 'POY',
                   'TOY', 'rXES', 'rXRF', 'rXEOL']  # All that are normalized to mesh

    # Note that the data dict only gets populated locally until the appropriate
    # singular y stream is return -- Assignment of y_stream happens after evaluation
    # of mathemtatical expression

    def get_y_data(y_stream, data, arg, background, REIXSObj):
        if doesMatchPattern(y_stream, special_streams):
            # CAUTION: Order matters as doesmatchpattern also returns TRUE if pattern in instance
            # is recogniced --> shortest pattern to check last

            if doesMatchPattern(y_stream, ['rXES']):
                roi = y_stream.lstrip("rXES[").rstrip("]")

                roi_low = float(roi.split(":")[0])
                roi_high = float(roi.split(":")[1])

                return data[arg].rXES(roi_low, roi_high)

            elif y_stream == 'XES':
                return data[arg].XES()

            elif doesMatchPattern(y_stream, ['rXRF']):
                roi = y_stream.lstrip("rXRF[").rstrip("]")

                roi_low = float(roi.split(":")[0])
                roi_high = float(roi.split(":")[1])
                return data[arg].rXRF(roi_low, roi_high)

            elif y_stream == 'XRF':
                return data[arg].XRF()

            elif y_stream == 'TEY':
                return data[arg].TEY

            elif y_stream == 'TFY':
                return data[arg].TFY()

            elif doesMatchPattern(y_stream, ["iPFY"]):
                roi = y_stream.lstrip("iPFY[").rstrip("]")

                if check_key_in_dict(roi, EdgeDict):
                    return data[arg].iPFY(iPFY_edge=roi)

                else:
                    roi_low = float(roi.split(":")[0])
                    roi_high = float(roi.split(":")[1])
                    return data[arg].iPFY(iSDDLowerBound=roi_low, iSDDUpperBound=roi_high)

            elif doesMatchPattern(y_stream, ["specPFY"]):
                roi = y_stream.lstrip("specPFY[").rstrip("]")

                roi_low = float(roi.split(":")[0])
                roi_high = float(roi.split(":")[1])
                return data[arg].specPFY(roi_low, roi_high)

            elif doesMatchPattern(y_stream, ["PFY"]):
                roi = y_stream.lstrip("PFY[").rstrip("]")

                if check_key_in_dict(roi, EdgeDict):
                    return data[arg].PFY(PFY_edge=roi)

                else:
                    roi_low = float(roi.split(":")[0])
                    roi_high = float(roi.split(":")[1])
                    return data[arg].PFY(SDDLowerBound=roi_low, SDDUpperBound=roi_high)

            elif y_stream == 'XEOL':
                return xeol_spec(data, arg, REIXSobj, background_scan=background)

            elif doesMatchPattern(y_stream, ["rXEOL"]):
                roi = y_stream.lstrip("rXEOL[").rstrip("]")
                roi_low = float(roi.split(":")[0])
                roi_high = float(roi.split(":")[1])

                return rxeol_spec(data, arg, REIXSobj, roi_low, roi_high, background_scan=background)

            elif y_stream == 'TOY':
                return toy_spec(data, arg, REIXSobj, background_scan=background)

            elif doesMatchPattern(y_stream, ["POY"]):
                roi = y_stream.lstrip("POY[").rstrip("]")
                roi_low = float(roi.split(":")[0])
                roi_high = float(roi.split(":")[1])

                return poy_spec(data, arg, REIXSobj, roi_low, roi_high, background_scan=background)

            elif y_stream == 'EY':
                return data[arg].sample_current

            # elif y_stream == 'tey':
            #    # spec mnemonic for sample current
            #    return data[arg].sample_current

            elif y_stream == 'Sample':
                # to ensure backwards compatibility
                return data[arg].sample_current

            elif y_stream == 'Mesh':
                # to ensure backwards compatibility
                return data[arg].mesh_current

            else:
                try:
                    return get_sca_data(y_stream, data, arg)

                except:
                    raise UserWarning("Special Stream not defined.")

        else:
            return get_sca_data(y_stream, data, arg)

    def get_sca_data(stream, data, arg):
        try:
            return np.array(data[arg].sca_data[stream])
        except:
            return np.array(data[arg].sca_data[data[arg].mnemonic2name[stream]])

    def get_x_data(x_stream, data, arg, background, REIXSObj):
        if x_stream == "Mono Energy":
            return data[arg].mono_energy

        elif x_stream == "MCP Energy":
            return data[arg].mcp_energy

        elif x_stream == "SDD Energy":
            return data[arg].sdd_energy

        elif x_stream == "XEOL Energy":
            return data[arg].xeol_energy

        elif x_stream == "Points":
            return np.array(range(0, len(data[arg].y_stream)), dtype=int)

        else:
            return get_sca_data(x_stream, data, arg)

    def grid_data(x_stream, y_stream, grid):
        xmin = grid[0]
        xmax = grid[1]

        numPoints = int(np.ceil((xmax-xmin)/grid[2])) + 1
        new_x = np.linspace(xmin, xmax, numPoints)
        f = interp1d(x_stream, y_stream, fill_value='extrapolate')
        new_y = f(new_x)

        return new_x, new_y

    data = dict()
    REIXSobj = REIXS(basedir, file)
    for arg in args:
        data[arg] = REIXSobj.Scan(arg)

        # Assign the calculated result to the y_stream of the object in data dict
        data[arg].y_stream = math_stream(
            y_stream, data, arg, get_y_data, XAS_streams=XAS_streams, is_XAS=is_XAS, background=background, REIXSObj=REIXSobj)

        # Assign the calculated result to the x_stream of the object in data dict
        data[arg].x_stream = math_stream(x_stream, data, arg, get_x_data)

        # Grid the data if specified
        if grid_x != [None, None, None]:
            new_x, new_y = grid_data(
                data[arg].x_stream, data[arg].y_stream, grid_x)

            data[arg].x_stream = new_x
            data[arg].y_stream = new_y

        # Apply offsets and normalize
        data[arg].x_stream = apply_offset(
            data[arg].x_stream, xoffset, xcoffset)

        if norm == True:
            data[arg].y_stream = np.interp(
                data[arg].y_stream, (data[arg].y_stream.min(), data[arg].y_stream.max()), (0, 1))

        data[arg].y_stream = apply_offset(
            data[arg].y_stream, yoffset, ycoffset)

        # Take derivatives
        if deriv != None:
            data[arg].y_stream = take_derivative1d(
                data[arg].y_stream, data[arg].x_stream, deriv)
            if norm == True:
                data[arg].y_stream = data[arg].y_stream / \
                    data[arg].y_stream.max()

        # Transforms RIXS to energy loss scale if incident energy is given
        if energyloss != None:
            data[arg].x_stream = energyloss-data[arg].x_stream

    return data



import numpy as np

#########################################################################################
def XEOL_background_removal(data,arg,REIXSobj,background_scan=None):
    
    if background_scan!=None:
        # 1 Sum the background
        background_data = REIXSobj.Scan(background_scan).xeol_data
        background_sum = background_data.sum(axis=0)

        # 2 Normalize to average background frame
        background_spec = np.true_divide(background_sum,len(background_data))

        # 3 Subtract the background from the actual spectrum
        xeol_data = data[arg].xeol_data
        # To do so, need to broadcast background into shape
        xeol_subtracted = np.subtract(xeol_data,background_spec)

        return xeol_subtracted

    else:
        return data[arg].xeol_data

def xeol_spec(data,arg,REIXSobj,background_scan=None):
    xeol_subtracted = XEOL_background_removal(data,arg,REIXSobj,background_scan)
    return np.sum(xeol_subtracted,axis=0)

def xeol_idx_lambda(data,arg,lambda_low,lambda_high):
    idx_low  = (np.abs(lambda_low  - data[arg].xeol_energy)).argmin()
    idx_high = (np.abs(lambda_high - data[arg].xeol_energy)).argmin()
    return idx_low,idx_high

        
def xeol_idx_energy(data,arg,energy_low,energy_high):
    idx_low  = (np.abs(energy_low  - data[arg].mono_energy)).argmin()
    idx_high = (np.abs(energy_high - data[arg].mono_energy)).argmin()
    return idx_low,idx_high


def rxeol_spec(data,arg,REIXSobj,energy_low,energy_high,background_scan=None):
    xeol_subtracted = XEOL_background_removal(data,arg,REIXSobj,background_scan)
    ixd_low,idx_high = xeol_idx_energy(data,arg,energy_low,energy_high)
    return np.sum(np.true_divide(xeol_subtracted[ixd_low:idx_high,:],data[arg].mesh_current[ixd_low:idx_high,None]),axis=0)

def toy_spec(data,arg,REIXSobj,background_scan=None):
    xeol_subtracted = XEOL_background_removal(data,arg,REIXSobj,background_scan)
    return np.sum(np.true_divide(xeol_subtracted,data[arg].mesh_current[:,None]),axis=1)

def poy_spec(data,arg,REIXSobj,lambda_low,lambda_high,background_scan=None):
    xeol_subtracted = XEOL_background_removal(data,arg,REIXSobj,background_scan)
    ixd_low,idx_high = xeol_idx_lambda(data,arg,lambda_low,lambda_high)
    return np.sum(np.true_divide(xeol_subtracted[:,ixd_low:idx_high],data[arg].mesh_current[:,None]),axis=1)


def xeol_map(data,arg,REIXSobj,background_scan=None):
    xeol_subtracted = XEOL_background_removal(data,arg,REIXSobj,background_scan)
    return np.true_divide(xeol_subtracted,data[arg].mesh_current[:,None])



import re
import parser
from .util import doesMatchPattern
from numpy import log as ln
from numpy import log10 as log
from numpy import exp


def math_stream(formula, data, arg, get_data, XAS_streams=None, is_XAS=False, background=None, REIXSObj=None):
    # Split the user input string at all mathematical operations
    # Allow "( ) * / + -" as math

    pattern = '[\(+\-*^/\)]'
    split_expr = re.split(pattern, formula)

    # Place string literals in dict if cannot be converted to float
    # drop all empty strings from re.split
    quantity_str_dict = dict()

    for i, string in enumerate(split_expr):
        if string != "":
            try:
                float(string)
            except:
                # Use math expressions to allow logs and exps
                math_expressions = ['ln', 'log', 'exp']

                if string in math_expressions:
                    pass

                else:
                    # Assign generic "val{i}" key to string literal in compliance with
                    # python supported syntax for variables
                    quantity_str_dict[f"val{i}"] = string

    # Parser does not support special string literals (ROIs) due to inproper python variable naming syntax
    # Replace them with generic key as per dictionary
    # Create local variable and assign corresponding data -- needed as eval interprets all input as variable
    # This is indeed local to this function only and cannot be accessed from loadSCAscans
    for k, v in quantity_str_dict.items():
        formula = formula.replace(v, k)

        # Ensure that in XASLoader all quantities are normalized by mesh
        # per definition of what XAS is
        # Exclude those quantities from normalization that have been normalized elsewhere

        if is_XAS == False:
            locals()[k] = get_data(v, data, arg, background, REIXSObj)
        else:
            if v == 'tey':
                # spec nmemonic tey means sample current - not yet normalized to mesh current
                numerator = get_data(v, data, arg, background, REIXSObj)
                mesh = data[arg].mesh_current
                locals()[k] = numerator/mesh

            elif doesMatchPattern(v, XAS_streams):
                locals()[k] = get_data(v, data, arg, background, REIXSObj)
            else:
                numerator = get_data(v, data, arg, background, REIXSObj)
                mesh = data[arg].mesh_current
                locals()[k] = numerator/mesh

    # Return the calculated result
    code = parser.expr(formula).compile()
    return eval(code)
